@@ -7,99 +7,103 @@ import time
 
 from datetime import datetime
 from configparser import ConfigParser
+from btrfs_sxbackup.SnapshotName import SnapshotName
+from btrfs_sxbackup.KeepExpression import KeepExpression
+from btrfs_sxbackup.Configuration import Configuration
 
 
 class Backup:
     """ Backup """
 
-    CONFIG_FILENAME = '/etc/btrfs-sxbackup.conf'
-
     class Error(Exception):
         pass
 
-    class Snapshot:
-        def __init__(self, name):
-            self.name = name
-
-    class Configuration:
-        """ btrfs-sxbackup configuration file """
-
-        __KEY_SOURCE = 'source'
-        __KEY_SOURCE_CONTAINER = 'source-container'
-        __KEY_DESTINATION = 'destination'
-        __KEY_KEEP = 'keep'
-
-        def __init__(self, location=None):
-            """
-            c'tor
-            :param location: Location this configuration refers to/resides or none for global configuration file
-            """
-
-            self.location = location
-            self.source = None
-            self.source_container = None
-            self.destination = None
-            self.keep = '1w = 2/d, 2w = daily, 1m = weekly, 2m = none'
-
-        @staticmethod
-        def __section_name_by_location(location):
-            """
-            Delivers the default configuration section name by location instance/type
-            :param location: Location instance or None
-            :return: Section name string
-            """
-            if isinstance(location, Backup.SourceLocation):
-                return 'Source'
-            else:
-                if isinstance(location, Backup.DestinationLocation):
-                    return 'Destination'
-                else:
-                    return 'Global'
-
-        def read(self, fileobject):
-            cparser = ConfigParser()
-            cparser.read_file(fileobject)
-
-            section_name = self.__section_name_by_location(self.location)
-            self.source = cparser.get(section_name, self.__KEY_SOURCE, fallback=None)
-            self.source_container = cparser.get(section_name, self.__KEY_SOURCE_CONTAINER, fallback=None)
-            self.destination = cparser.get(section_name, self.__KEY_DESTINATION, fallback=None)
-            self.keep = cparser.get(section_name, self.__KEY_KEEP, fallback=self.keep)
-
-        def write(self, fileobject):
-            cparser = ConfigParser()
-
-            section_name = self.__section_name_by_location(self.location)
-            cparser.add_section(section_name)
-            if self.source is not None:
-                cparser.set(section_name, self.__KEY_SOURCE, self.source)
-            if self.source_container is not None:
-                cparser.set(section_name, self.__KEY_SOURCE_CONTAINER, self.source_container)
-            if self.destination is not None:
-                cparser.set(section_name, self.__KEY_DESTINATION, self.destination)
-            if self.keep is not None:
-                cparser.set(section_name, self.__KEY_KEEP, self.keep)
-            cparser.write(fileobject)
-
     class Location:
         """ Backup location """
 
         __TEMP_BACKUP_NAME = 'temp'
+        __CONFIG_FILENAME = '.btrfs-sxbackup'
 
-        def __init__(self, url, container_subvolume, max_snapshots):
+        class Configuration:
+            """ btrfs-sxbackup configuration file """
+
+            __KEY_SOURCE = 'source'
+            __KEY_SOURCE_CONTAINER = 'source-container'
+            __KEY_DESTINATION = 'destination'
+            __KEY_KEEP = 'keep'
+
+            def __init__(self, location=None):
+                """
+                c'tor
+                :param location: Location this configuration refers to/resides or none for global configuration file
+                """
+
+                self.location = location
+                self.source = None
+                self.source_container = None
+                self.destination = None
+                self.keep = KeepExpression('1w = 2/d, 2w = daily, 1m = weekly, 2m = none')
+
+            @staticmethod
+            def __section_name_by_location(location):
+                """
+                Delivers the default configuration section name by location instance/type
+                :param location: Location instance or None
+                :return: Section name string
+                """
+                if isinstance(location, Backup.SourceLocation):
+                    return 'Source'
+                else:
+                    if isinstance(location, Backup.DestinationLocation):
+                        return 'Destination'
+                    else:
+                        raise ValueError('Configuration does not support location instance type [%s]' % location)
+
+            def read(self, fileobject):
+                cparser = ConfigParser()
+                cparser.read_file(fileobject)
+
+                section_name = self.__section_name_by_location(self.location)
+                self.source = cparser.get(section_name, self.__KEY_SOURCE, fallback=None)
+                self.source_container = cparser.get(section_name, self.__KEY_SOURCE_CONTAINER, fallback=None)
+                self.destination = cparser.get(section_name, self.__KEY_DESTINATION, fallback=None)
+                self.keep = cparser.get(section_name, self.__KEY_KEEP, fallback=self.keep)
+
+            def write(self, fileobject):
+                cparser = ConfigParser()
+
+                section_name = self.__section_name_by_location(self.location)
+                cparser.add_section(section_name)
+                if self.source is not None:
+                    cparser.set(section_name, self.__KEY_SOURCE, self.source)
+                if self.source_container is not None:
+                    cparser.set(section_name, self.__KEY_SOURCE_CONTAINER, self.source_container)
+                if self.destination is not None:
+                    cparser.set(section_name, self.__KEY_DESTINATION, self.destination)
+                if self.keep is not None:
+                    cparser.set(section_name, self.__KEY_KEEP, self.keep)
+                cparser.write(fileobject)
+
+        def __init__(self, url, container_subvolume, keep):
+            """
+            c'tor
+            :param url: Location URL
+            :param container_subvolume: Subvolume name
+            :param keep: Keep expression instance
+            """
             self.__logger = logging.getLogger(self.__class__.__name__)
             self.url = url
             self.container_subvolume = os.path.join(url.path, container_subvolume)
-            self.max_snapshots = max_snapshots
+            self.keep = keep
             self.snapshot_names = []
-            self.configuration_filename = os.path.join(self.container_subvolume, '.btrfs-sxbackup')
+            self.configuration_filename = os.path.join(self.container_subvolume, self.__CONFIG_FILENAME)
 
             # Path of subvolume for current backup run
             # Subvolumes will be renamed from temp to timestamp based name on both side if successful.
             self.temp_subvolume = os.path.join(self.container_subvolume, self.__TEMP_BACKUP_NAME)
 
             # Override configuration params
-            self.configuration = Backup.Configuration(self)
+            self.configuration = Backup.Location.Configuration(self)
 
         def __format_log_msg(self, msg):
             return '%s :: %s' % (self.get_name(), msg)
@@ -118,7 +122,8 @@ class Backup:
 
         def create_subprocess_args(self, cmd):
             """
-            Create command/args array for subprocess, wraps command into ssh call if url host name is not None
+            Create subprocess arguments for shell command/args to be executed in this location.
+            Internally Wraps command into ssh call if url host name is not None
             :param cmd: Shell command
             :return: Subprocess arguments
             """
@@ -136,7 +141,7 @@ class Backup:
             """ Creates bash comand string to remove multiple snapshots within a btrfs subvolume """
 
             return " && ".join(
-                map(lambda x: 'btrfs sub del %s' % (os.path.join(self.container_subvolume, x)), snapshot_names))
+                map(lambda x: 'btrfs sub del %s' % (os.path.join(self.container_subvolume, str(x))), snapshot_names))
 
         def prepare_environment(self):
             """ Prepare location environment """
@@ -154,28 +159,31 @@ class Backup:
                 self.create_subprocess_args('btrfs sub list -o %s' % self.container_subvolume))
             # output is delivered as a byte sequence, decode to unicode string and split lines
             lines = output.decode().splitlines()
-            # extract snapshot names from btrfs sub list lines
 
+            # extract snapshot names from btrfs sub list lines
             def strip_name(l):
                 i = l.rfind(os.path.sep)
                 return l[i + 1:] if i >= 0 else l
 
             lines = map(lambda x: strip_name(x), lines)
+
             # sort and return
-            self.snapshot_names = sorted(lines, reverse=True)
+            snapshot_names = map(lambda sn: SnapshotName(sn), lines)
+            self.snapshot_names = sorted(snapshot_names, key=lambda sn: sn.timestamp)
             return self.snapshot_names
 
         def cleanup_snapshots(self):
             """ Clean out excess backups/snapshots """
+            if self.keep is not None:
+                (to_remove_by_condition, to_keep) = self.keep.filter(self.snapshot_names, lambda sn: sn.timestamp)
 
-            if len(self.snapshot_names) > self.max_snapshots:
-                remove_count = len(self.snapshot_names) - self.max_snapshots
-                snapshots_to_remove = self.snapshot_names[-remove_count:]
-                # self.log_info('Removing snapshots [%s]' % (", ".join(snapshots_to_remove)))
-                self.log_info('Removing %d of %d snapshots (because >%d): %s'
-                              % (remove_count, len(self.snapshot_names), self.max_snapshots, snapshots_to_remove))
-                subprocess.check_output(
-                    self.create_subprocess_args(self.create_cleanup_bash_command(snapshots_to_remove)))
+                for c in to_remove_by_condition.keys():
+                    to_remove = to_remove_by_condition[c]
+
+                    self.log_info('Removing %d snapshot(s) due to condition %s: %s'
+                                  % (len(to_remove), str(c), list(map(lambda x: str(x), to_remove))))
+                    subprocess.check_output(
+                        self.create_subprocess_args(self.create_cleanup_bash_command(to_remove)))
 
         def write_configuration(self):
             """ Write configuration file to container subvolume """
@@ -227,12 +235,22 @@ class Backup:
         def get_name(self):
             return 'Destination'
 
-    def __init__(self, source_url, source_container_subvolume, source_max_snapshots, dest_url, dest_max_snapshots):
-        """ c'tor """
+    def __init__(self, config, source_url, source_container_subvolume, source_keep, dest_url, dest_keep):
+        """
+        c'tor
+        :param config: Global configuration instance
+        :param source_url: Source URL
+        :param source_container_subvolume: Source container subvolume name
+        :param source_keep: Source keep expression instance
+        :param dest_url: Destination URL
+        :param dest_keep: Destination keep expression instance
+        :return:
+        """
         self.__logger = logging.getLogger(self.__class__.__name__)
 
-        self.source = Backup.SourceLocation(source_url, source_container_subvolume, source_max_snapshots)
-        self.dest = Backup.DestinationLocation(dest_url, "", dest_max_snapshots)
+        self.config = config
+        self.source = Backup.SourceLocation(source_url, source_container_subvolume, source_keep)
+        self.dest = Backup.DestinationLocation(dest_url, "", dest_keep)
 
         self.compress = False
 
@@ -264,10 +282,6 @@ class Backup:
         self.__logger.info(self.dest)
 
         starting_time = time.monotonic()
-        # Read global configuration
-        config = Backup.Configuration()
-        if os.path.exists(Backup.CONFIG_FILENAME):
-            config.read(open(Backup.CONFIG_FILENAME))
 
         # Prepare environments
         self.__logger.info('Preparing environment')
@@ -289,6 +303,7 @@ class Backup:
         self.source.configuration.destination = \
             self.dest.url.geturl() if self.dest.is_remote() or both_remote_or_local \
             else None
+        self.source.configuration.keep = self.source.keep.expression_text if self.source.keep is not None else None
 
         self.dest.configuration.source = \
             self.source.url.geturl() if self.source.is_remote() or both_remote_or_local \
@@ -299,16 +314,18 @@ class Backup:
         self.dest.configuration.destination = \
             self.dest.url.geturl() if both_remote_or_local \
             else None
+        self.dest.configuration.keep = self.dest.keep.expression_text if self.dest.keep is not None else None
 
         # Retrieve snapshot names of both source and destination 
         self.source.retrieve_snapshot_names()
         self.dest.retrieve_snapshot_names()
 
-        new_snapshot_name = self.__create_snapshot_name()
-        if len(self.source.snapshot_names) > 0 and new_snapshot_name <= self.source.snapshot_names[0]:
+        new_snapshot_name = SnapshotName()
+        if len(self.source.snapshot_names) > 0 \
+                and new_snapshot_name.timestamp <= self.source.snapshot_names[0].timestamp:
             raise Backup.Error('Current snapshot name [%s] would be older than newest existing snapshot [%s] \
                                  which may indicate a system time problem'
-                                 % (new_snapshot_name, self.source.snapshot_names[0]))
+                               % (new_snapshot_name, self.source.snapshot_names[0]))
 
         # Create source snapshot
         self.source.create_snapshot()
@@ -321,7 +338,7 @@ class Backup:
             send_command_str = 'btrfs send %s' % self.source.temp_subvolume
         else:
             send_command_str = 'btrfs send -p %s %s' % (
-                os.path.join(self.source.container_subvolume, self.source.snapshot_names[0]),
+                os.path.join(self.source.container_subvolume, str(self.source.snapshot_names[0])),
                 self.source.temp_subvolume)
 
         if self.compress:
@@ -352,11 +369,17 @@ class Backup:
             send_returncode = send_process.poll()
 
             if receive_returncode is not None and receive_returncode != 0:
-                send_process.kill()
+                try:
+                    send_process.kill()
+                except ProcessLookupError:
+                    pass
                 break
 
             if send_returncode is not None and send_returncode != 0:
-                receive_process.kill()
+                try:
+                    receive_process.kill()
+                except ProcessLookupError:
+                    pass
                 break
 
             time.sleep(2)
@@ -375,9 +398,9 @@ class Backup:
         self.__logger.info('Finalizing backup')
         subprocess.check_output(self.source.create_subprocess_args(
             'mv %s %s' % (
-                self.source.temp_subvolume, os.path.join(self.source.container_subvolume, new_snapshot_name))))
+                self.source.temp_subvolume, os.path.join(self.source.container_subvolume, str(new_snapshot_name)))))
         subprocess.check_output(self.dest.create_subprocess_args(
-            'mv %s %s' % (self.dest.temp_subvolume, os.path.join(self.dest.url.path, new_snapshot_name))))
+            'mv %s %s' % (self.dest.temp_subvolume, os.path.join(self.dest.url.path, str(new_snapshot_name)))))
 
         # Update snapshot name lists
         self.source.snapshot_names = [new_snapshot_name] + self.source.snapshot_names
