@@ -14,7 +14,7 @@ class Configuration:
     __key_EMAIL_RECIPIENT = 'email-recipient'
 
     def __init__(self):
-        self.keep = KeepExpression('1w = 2/d, 2w = daily, 1m = weekly, 2m = none')
+        self.keep = None
         self.log_ident = None
         self.email_recipient = None
 
@@ -25,6 +25,7 @@ class Configuration:
             with open(self.__CONFIG_FILENAME, 'r') as file:
                 cparser.read_file(file)
 
-            self.keep = KeepExpression(cparser.get(self.__SECTION_NAME, self.__KEY_KEEP, fallback=self.keep))
+            keep_str = cparser.get(self.__SECTION_NAME, self.__KEY_KEEP, fallback=self.keep)
+            self.keep = KeepExpression(keep_str) if keep_str else None
             self.log_ident = cparser.get(self.__SECTION_NAME, self.__KEY_LOG_IDENT, fallback=None)
             self.email_recipient = cparser.get(self.__SECTION_NAME, self.__key_EMAIL_RECIPIENT, fallback=None)
